// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

void nvfs_free_blocks(struct nvfs_superblock *nvs, unsigned long i, unsigned long n)
{
	unsigned long group, offset;
	uint8_t *bitmap;
	uint8_t *flush_start, *flush_end;
	struct nvfs_group *g;
	struct nvfs_group_info *gi;
	unsigned long nn = n;

next:
	group = i >> log2_group_size(nvs);
	offset = (i & ((1UL << log2_group_size(nvs)) - 1)) >> nvs->log2_page_size;

	if (unlikely(group >= nvs->n_groups)) {
		nvfs_error(nvs, "freeing invalid block %lu", i);
		return;
	}

	bitmap = p_bitmap(nvs, group);
	gi = &nvs->group_info[group];

	nv_log_suppress(nvs);
	spin_lock(&gi->lock);

	flush_start = bitmap + offset / 8;

next_in_group:
	if (!__test_and_clear_bit_le(offset, bitmap)) {
		nvfs_error(nvs, "block bit %lu already cleared", i);
		spin_unlock(&gi->lock);
		nv_log_resume(nvs);
		return;
	}

	g = p_group(nvs, group);

	gi->n_free_blocks++;

	n--;
	i += 1UL << nvs->log2_page_size;
	if (n && likely(offset != ((1UL << log2_group_size(nvs)) - 1) >> nvs->log2_page_size)) {
		/* !!! TODO: spin_is_contended is dummy on most architectuers */
		if (unlikely(spin_is_contended(&gi->lock)) && unlikely(!(n & 0xf)))
			goto drop_spinlock;
		offset++;
		goto next_in_group;
	}

drop_spinlock:
	flush_end = bitmap + offset / 8;
	nv_flush(nvs, flush_start, flush_end - flush_start + 1);

	nv_store_cached(nvs, g->n_free_blocks, cpu_to_le32(gi->n_free_blocks));

	spin_unlock(&gi->lock);
	nv_log_resume(nvs);

	cond_resched();

	if (unlikely(n != 0))
		goto next;

	percpu_counter_add(&nvs->free_blocks_cnt, nn);
}

unsigned long nvfs_free_run(struct nvfs_superblock *nvs, __le64 *run, unsigned long n, bool zero)
{
	/*
	 * TODO: optimization - take the lock just one when freeing
	 * in the same group
	 */
	unsigned long freed = 0;
	while (n--) {
		unsigned long to_free = 1;
		unsigned long block_to_free = le64_to_cpu(*run);
		if (block_to_free) {
			while (to_free <= n && le64_to_cpu(run[to_free]) == block_to_free + (to_free << nvs->log2_page_size)) {
				to_free++;
				cond_resched();
			}
			freed += to_free;
			/*printk("free: %lx, %lx\n", block_to_free, to_free);*/
			nvfs_free_blocks(nvs, block_to_free, to_free);
			if (zero) {
				unsigned long i;
				for (i = 0; i < to_free; i++) {
					nv_store(nvs, run[i], cpu_to_le64(0));
					cond_resched();
				}
			}
			n -= to_free - 1;
		}
		run += to_free;
		cond_resched();
	}
	return freed;
}

static unsigned long block_try_group(struct nvfs_superblock *nvs, unsigned long group, unsigned long offset, unsigned long limit, unsigned *prealloc)
{
	struct nvfs_group *g;
	struct nvfs_group_info *gi;
	uint8_t *bitmap;
	unsigned long o;

	if (unlikely(WARN_ON(group >= nvs->n_groups)))
		return ~0;

	g = p_group(nvs, group);
	gi = &nvs->group_info[group];
	bitmap = p_bitmap(nvs, group);

	if (unlikely(READ_ONCE(g->n_free_blocks) < limit))
		return ~0;

search_again:
	o = nvfs_find_bit_in_bmp(bitmap, offset, 1UL << log2_pages_per_group(nvs));
	if (likely(o != ~0)) {
		unsigned n_allocated;
		unsigned long flush_start, flush_end;

		nv_log_suppress(nvs);
		spin_lock(&gi->lock);

		if (unlikely(__test_and_set_bit_le(o, bitmap))) {
			/* we have lost the race */
			spin_unlock(&gi->lock);
			nv_log_resume(nvs);
			goto search_again;
		}

		n_allocated = 1;
		if (prealloc) {
			unsigned p;
			unsigned long limit = o + *prealloc;
			if (limit > 1UL << log2_pages_per_group(nvs))
				limit = 1UL << log2_pages_per_group(nvs);
			for (p = o + 1; p < limit; p++)
				if (__test_and_set_bit_le(p, bitmap))
					break;
			n_allocated = *prealloc = p - o;
		}

		flush_start = o / 8;
		flush_end = (o + n_allocated + 7) / 8;
		nv_flush(nvs, bitmap + flush_start, flush_end - flush_start);

		if (unlikely(o == 1) || (unlikely(!o) && unlikely(!group))) {
			spin_unlock(&gi->lock);
			nv_log_resume(nvs);
			nvfs_error(nvs, "allocating reserved block %lu", (group << log2_pages_per_group(nvs)) | o);
			return ~0;
		}

		gi->n_free_blocks -= n_allocated;
		nv_store_cached(nvs, g->n_free_blocks, cpu_to_le32(gi->n_free_blocks));
		nv_wmb(nvs);
		spin_unlock(&gi->lock);
		nv_log_resume(nvs);
		percpu_counter_sub(&nvs->free_blocks_cnt, n_allocated);
	}

	return o;
}

uint64_t nvfs_new_block(struct nvfs_superblock *nvs, unsigned *prealloc)
{
	unsigned long group, offset, o, i, next;
	struct nvfs_per_cpu *npc = raw_cpu_ptr(nvs->npc);

	i = READ_ONCE(npc->current_block);
	group = i >> log2_pages_per_group(nvs);
	offset = i & ((1UL << log2_pages_per_group(nvs)) - 1);

	cond_resched();
	o = block_try_group(nvs, group, offset, 1, prealloc);
	if (unlikely(o == ~0)) {
		unsigned long x;
		uint64_t freei = percpu_counter_read_positive(&nvs->free_blocks_cnt);
		uint64_t average;

		average = div64_u64(freei, nvs->n_groups);

		offset = nvfs_get_random(1UL << log2_pages_per_group(nvs));
		for (x = 0; x < nvs->n_groups; x++) {
			cond_resched();
			group = nvfs_get_random(nvs->n_groups);
			o = block_try_group(nvs, group, offset, average, prealloc);
			if (o != ~0)
				goto found;
		}
		for (group = 0; group < nvs->n_groups; group++) {
			cond_resched();
			o = block_try_group(nvs, group, offset, 1, prealloc);
			if (o != ~0)
				goto found;
		}
		return 0;
	}

found:
	i = (group << log2_pages_per_group(nvs)) | o;

	if (prealloc)
		next = i + *prealloc;
	else
		next = i + 1;
	if (unlikely(next < nvs->n_groups << log2_pages_per_group(nvs)))
		WRITE_ONCE(npc->current_block, next);
	else
		WRITE_ONCE(npc->current_block, 0);

	return i << nvs->log2_page_size;
}
